SystemOrganization addCategory: #'GT-DebuggerExtensions'!

!Context methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 1/5/2015 16:53'!
isHaltContext

	^ self method selector = #halt! !

!DebugSession methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/26/2013 11:40'!
createPredicateForBlock: aBlock 

	^ BlockContextPredicate for: aBlock! !

!DebugSession methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/14/2013 18:17'!
createStepPredicateForBlock: aBlock 

	^ self createStepPredicatForPredicate: (self createPredicatForBlock: aBlock)
			! !

!DebugSession methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/20/2013 16:46'!
createStepPredicateForPredicate: aPredicate
	" Subclasses should specialize this method if needed. "
	^ aPredicate! !

!DebugSession methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/20/2013 17:22'!
createStepPredicateForPredicate: aPredicate fromContext: targetContext

	^ (self createPredicateForBlock: [ :currentContext |
		(self isActive: targetContext) not or: [ currentContext isHaltContext ] ]) 
			or: aPredicate
	! !

!DebugSession methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/26/2013 11:51'!
stepUntil: aPredicate from: aContext 
	| ctxt |
	
	ctxt := self process stepUntil: (self createStepPredicateForPredicate: aPredicate) from: aContext.
	self  
		updateWithContext: ctxt 
		fromProcess: self process.! !

!DummyUIManager methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 9/17/2013 19:01'!
handleMoldableHalt: anException

	self unhandledErrorDefaultAction: anException! !

!CommandLineUIManager methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 9/17/2013 19:01'!
handleMoldableHalt: anException

	self unhandledErrorDefaultAction: anException

! !

!GTMoldableDebugger class methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 3/14/2016 15:18'!
openOn: aDebugSession withFullView: aBool notification: aString initializedWith: aBlock
	
	| debugger |
	
	debugger := self on: aDebugSession.
	
	aBool | self alwaysOpenFullDebugger
		ifTrue: [ debugger openWithFullView ]
		ifFalse: [ debugger openWithNotification: aString].
	aBlock value: debugger.! !

!ToolRegistry methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 9/17/2013 19:01'!
debugMoldableHalt: anException
	^Processor activeProcess
		debugMoldableHalt: anException signalerContext sender "TODO: calling 'sender' might not be the best indea "
		title: anException description! !

!ToolRegistry methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 11/27/2015 18:56'!
debuggerDispatcherFor: context
	^ self debuggerDispatcherFor: context matching: [ :aDebugger | true ].
	! !

!InstructionStream methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 6/29/2013 20:29'!
isInitialnstruction

	^ self pc = self method initialPC! !

DebugAction subclass: #GTReflectivityDebugAction
	instanceVariableNames: 'breakpoints debuggerData'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!GTReflectivityDebugAction methodsFor: 'adding' stamp: 'AndreiChis 3/14/2016 14:11'!
addBreakpoint: aBreakpoint

	aBreakpoint targetAction: self.
	self breakpoints add: aBreakpoint! !

!GTReflectivityDebugAction methodsFor: 'callbacks' stamp: 'AndreiChis 9/6/2017 09:43'!
attachBreakpoints
	self subclassResponsibility! !

!GTReflectivityDebugAction methodsFor: 'callbacks' stamp: 'AndreiChis 9/7/2017 10:25'!
breakpointReachedFrom: aBreakpoint
	"Callback method called when any one of the breakpoints installed 
	by this action is reached."

	self uninstallBreakpoints.
	self openDebuggerForBreakpoint: aBreakpoint! !

!GTReflectivityDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 14:27'!
breakpoints

	^ breakpoints ifNil: [ breakpoints := OrderedCollection new ]! !

!GTReflectivityDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 9/7/2017 10:21'!
debuggerClass
	self subclassResponsibility! !

!GTReflectivityDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 15:13'!
debuggerData

	^ debuggerData ifNil: [ debuggerData := Dictionary new ]! !

!GTReflectivityDebugAction methodsFor: 'actions' stamp: 'AndreiChis 3/14/2016 14:08'!
executeAction

	self attachBreakpoints.
	self installBreakpoints.
	self resumeDebugger.! !

!GTReflectivityDebugAction methodsFor: 'initialization' stamp: 'AndreiChis 3/14/2016 14:10'!
initialize
	super initialize.
	self needsUpdate: false! !

!GTReflectivityDebugAction methodsFor: 'actions' stamp: 'AndreiChis 9/7/2017 10:22'!
installBreakpoints
	self breakpoints do: #install.! !

!GTReflectivityDebugAction methodsFor: 'opening' stamp: 'AndreiChis 9/7/2017 22:35'!
openDebuggerForBreakpoint: aBreakpoint
	|ctxt|
	ctxt := thisContext.
	
	UIManager default 
		debugProcess: Processor activeProcess 
		context: (aBreakpoint instrumentationContextFrom: ctxt)
		label: (self debuggerData at: #title)
		fullView: true 
		notification: nil 
		usingDebugger: self debuggerClass
		initializedWith: [ :aDebugger | 
			"FIX ME: do not update the entire debugger. Instead have a hook for updating the 
			session before opening the debugger and one after to update the actual debugger."
			aBreakpoint updateDebuggingSession: aDebugger session.
			aDebugger updateBrowser. 
			aDebugger window 
				position: (self debuggerData at: #position);
				extent: (self debuggerData at: #extent) ]! !

!GTReflectivityDebugAction methodsFor: 'actions' stamp: 'AndreiChis 3/14/2016 14:04'!
resumeDebugger

	self saveDebuggerData.
	self session 
		resume;
		clear.
	self debugger close. ! !

!GTReflectivityDebugAction methodsFor: 'actions' stamp: 'AndreiChis 3/14/2016 15:14'!
saveDebuggerData

	self debuggerData at: #title put: self debugger title.
	self debuggerData at: #position put: self debugger window position.
	self debuggerData at: #extent put: self debugger window extent.! !

!GTReflectivityDebugAction methodsFor: 'actions' stamp: 'AndreiChis 9/7/2017 10:21'!
uninstallBreakpoints
	self breakpoints do: #uninstall.! !

!UIManager methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 9/17/2013 19:01'!
handleMoldableHalt: anException

	"See UIManager>>unhandledErrorDefaultAction:"
		
	Smalltalk snapshot: false andQuit: true.! !

!MorphicUIManager methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 3/14/2016 12:25'!
debugProcess: process context: context label: title fullView: bool notification: notificationString usingDebugger: debuggerClass initializedWith: aBlock
	"Open a notifier in response to an error, halt, or notify. A notifier view
	just shows a short view of the sender stack and provides a menu that
	lets the user open a full debugger."

	| debugSession |

	debugSession := debuggerClass sessionClass named: title on: process startedAt: context.
	
	debugSession logStackToFileIfNeeded.
	debugSession isAboutUIProcess ifTrue: [ self spawnNewProcess  ].
		
	self defer:	[ 
			"schedule debugger in deferred UI message to address
			redraw problems after opening a debugger e.g. from
			the testrunner."
		[ debuggerClass 
				openOn: debugSession withFullView: bool notification: notificationString initializedWith: aBlock.
		] on: Error do: [ :ex | 	debugSession signalDebuggerError: ex] 
	].

	process suspend.
! !

!MorphicUIManager methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 9/17/2013 19:01'!
handleMoldableHalt: anException

	^ Smalltalk tools debugMoldableHalt: anException.! !

!MorphicUIManager methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 1/26/2015 23:01'!
openDebugger: debuggerClass initializedWith: aBlock on: process context: context label: title contents:  contentsStringOrNil fullView: bool 
	"Open a notifier in response to an error, halt, or notify. A notifier view
	just shows a short view of the sender stack and provides a menu that
	lets the user open a full debugger."
	| errorWasInUIProcess  debugger |
	errorWasInUIProcess := self spawnNewProcessIfThisIsUI: process.
	[Smalltalk tools debugger logDebuggerStackToFile
		ifTrue: [Smalltalk
				logError: title
				inContext: context ]] on: Error do: [].
	debugger := debuggerClass new.
	self defer: [[ 
			debugger
				process: process
				controller: nil
				context: context.
			aBlock cull: debugger.
			"schedule debugger in deferred UI message to address
			redraw problems after opening a debugger e.g. from
			the testrunner."
			"self defer: ["
			bool
				ifTrue: [debugger openFullNoSuspendLabel: title]
				ifFalse: [debugger openNotifierContents: contentsStringOrNil label: title].
			debugger errorWasInUIProcess: errorWasInUIProcess.]
				on: Error
				do: [:ex | Smalltalk tools debugger primitiveError: 'Orginal error: ' , title asString , '.
	Smalltalk tools debugger error: '
							, ([ex description]
									on: Error
									do: ['a ' , ex class printString]) , ':']].

	process suspend.
	^ debugger! !

GTDebuggerContextPredicate subclass: #AndContextPredicate
	instanceVariableNames: 'left right'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!AndContextPredicate class methodsFor: 'instance creation' stamp: 'AndreiChis 7/17/2013 22:23'!
with: aPredicate with: anotherPredicate

	^ self new with: aPredicate with: anotherPredicate! !

!AndContextPredicate methodsFor: 'testing' stamp: 'AndreiChis 7/17/2013 22:24'!
matchContext: aContext

	^ (left matches: aContext ) and: [ right matches: aContext ]! !

!AndContextPredicate methodsFor: 'instance creation' stamp: 'AndreiChis 7/17/2013 22:33'!
with: aPredicate with: anotherPredicate

	left := aPredicate.
	right := anotherPredicate! !

GTDebuggerContextPredicate subclass: #BlockContextPredicate
	instanceVariableNames: 'blockCondition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!BlockContextPredicate commentStamp: '<historical>' prior: 0!
I encapsulate a condition that can be verified against a MethodContext as a block.
The block can have at least one parameter, representing the context that should be verified and 
should return a boolean value.
!

!BlockContextPredicate class methodsFor: 'instance creation' stamp: 'AndreiChis 7/17/2013 20:15'!
for: aBlock

	^ self new for: aBlock! !

!BlockContextPredicate methodsFor: 'instance creation' stamp: 'AndreiChis 7/17/2013 20:15'!
for: aBlock

	blockCondition := aBlock! !

!BlockContextPredicate methodsFor: 'testing' stamp: 'AndreiChis 7/17/2013 20:10'!
matchContext: aContext

	^ blockCondition isNil 
		ifFalse: [ blockCondition cull: aContext ] 
		ifTrue: [ false ]! !

!GTDebuggerContextPredicate methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/26/2013 11:39'!
and: aPredicate

	^ AndContextPredicate 
		with: self
		with: aPredicate! !

!GTDebuggerContextPredicate methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/26/2013 11:40'!
or: aPredicate

	^ OrContextPredicate 
		with: self
		with: aPredicate! !

GTDebuggerContextPredicate subclass: #OrContextPredicate
	instanceVariableNames: 'left right'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!OrContextPredicate class methodsFor: 'instance creation' stamp: 'AndreiChis 7/17/2013 22:23'!
with: aPredicate with: anotherPredicate

	^ self new with: aPredicate with: anotherPredicate! !

!OrContextPredicate methodsFor: 'testing' stamp: 'AndreiChis 7/17/2013 22:24'!
matchContext: aContext

	^ (left matches: aContext ) or: [ right matches: aContext ]! !

!OrContextPredicate methodsFor: 'instance creation' stamp: 'AndreiChis 7/17/2013 22:33'!
with: aPredicate with: anotherPredicate

	left := aPredicate.
	right := anotherPredicate! !

Object subclass: #GTDebuggerBrekpoint
	instanceVariableNames: 'metaLink condition conditionArguments targetAction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!GTDebuggerBrekpoint methodsFor: 'actions' stamp: 'AndreiChis 9/7/2017 10:31'!
breakpointReached
	"I am called when the breakpoint is reached. If there is an 
	associated condition with the breakpoint I am only called if
	the conditions holds."
	
	targetAction breakpointReachedFrom: self ! !

!GTDebuggerBrekpoint methodsFor: 'actions' stamp: 'AndreiChis 3/14/2016 14:31'!
install
	self subclassResponsibility! !

!GTDebuggerBrekpoint methodsFor: 'accessing' stamp: 'AndreiChis 9/7/2017 15:14'!
instrumentationContextFrom: aContext
	"The action of opening the debugger will be called from the debugging action.
	Hence the top context will not be the one where the metalink was declared,
	but the one were the debugger was opened. To avoid that we give the breakpoint a
	chance to return the correct context that contains the metalink instrumentation."
	
	^ aContext! !

!GTDebuggerBrekpoint methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 11:37'!
metaLink
	^ metaLink! !

!GTDebuggerBrekpoint methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 11:47'!
metaLink: anObject
	metaLink := anObject! !

!GTDebuggerBrekpoint methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 11:37'!
targetAction
	^ targetAction! !

!GTDebuggerBrekpoint methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 11:37'!
targetAction: anObject
	targetAction := anObject! !

!GTDebuggerBrekpoint methodsFor: 'actions' stamp: 'AndreiChis 3/14/2016 14:31'!
uninstall
	self metaLink uninstall ! !

!GTDebuggerBrekpoint methodsFor: 'actions' stamp: 'AndreiChis 9/7/2017 10:30'!
updateDebuggingSession: aSession
	"I am a callback that is called before the debugger is opened.
	I can be used to remove the instrumentation from the displayed methods.
	At this point the debugger process is suspended."! !

!GTDebuggerBrekpoint methodsFor: 'conditions' stamp: 'AndreiChis 3/14/2016 14:40'!
whenReceiver: aBlock

	condition := aBlock.
	conditionArguments := #(receiver)! !

GTDebuggerBrekpoint subclass: #GTDebuggerMethodAfterBreakpoint
	instanceVariableNames: 'targetMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!GTDebuggerMethodAfterBreakpoint methodsFor: 'actions' stamp: 'AndreiChis 3/14/2016 14:42'!
install

	self metaLink: (MetaLink new
		metaObject: self;
		selector: #breakpointReached;
		condition: condition arguments: conditionArguments;
		control: #after).
	targetMethod ast link: self metaLink.
! !

!GTDebuggerMethodAfterBreakpoint methodsFor: 'accessing' stamp: 'AndreiChis 9/7/2017 22:23'!
instrumentationContextFrom: aContext

	^ aContext  "--(GTReflectivityDebugAction)>>openDebuggerForBreakpoint:" 
		sender   "--(GTReflectivityDebugAction)>>breakpointReachedFrom:"
		sender   "--(GTDebuggerBrekpoint)>>breakpointReached"
		sender   "instrumented method -> ensure block calling breakpointReached"
		sender   "Context>>#resume:through:"
		sender   "Context>>ensure:"
		sender   "instrumented method -> before exiting"! !

!GTDebuggerMethodAfterBreakpoint methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 11:32'!
targetMethod
	^ targetMethod! !

!GTDebuggerMethodAfterBreakpoint methodsFor: 'accessing' stamp: 'AndreiChis 3/14/2016 11:32'!
targetMethod: anObject
	targetMethod := anObject! !

!GTDebuggerMethodAfterBreakpoint methodsFor: 'actions' stamp: 'AndreiChis 9/7/2017 22:30'!
updateDebuggingSession: aSession
	super updateDebuggingSession: aSession.
	
	aSession stepOver. "instrumented method -> ensure block calling breakpointReached"
	aSession stepOver. "ReflectiveMethod>>run:with:in:"! !

GTDebuggerBrekpoint subclass: #GTDebuggerMethodEntryBreakpoint
	instanceVariableNames: 'targetMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!GTDebuggerMethodEntryBreakpoint methodsFor: 'actions' stamp: 'AndreiChis 9/7/2017 09:27'!
install

	self metaLink: (MetaLink new
		metaObject: self;
		selector: #breakpointReached;
		condition: condition arguments: conditionArguments;
		control: #before).
	targetMethod ast link: self metaLink.
! !

!GTDebuggerMethodEntryBreakpoint methodsFor: 'accessing' stamp: 'AndreiChis 9/7/2017 18:21'!
instrumentationContextFrom: aContext

	^ aContext  "GTPPStepToNextProductionReflectivityAction(GTReflectivityDebugAction)>>openDebuggerForBreakpoint:"
		sender   "GTPPStepToNextProductionReflectivityAction(GTReflectivityDebugAction)>>breakpointReachedFrom:"
		sender   "GTDebuggerMethodEntryBreakpoint(GTDebuggerBrekpoint)>>breakpointReached"
		sender   "Instrumented method -> at the beginning after sending breakpointReached"! !

!GTDebuggerMethodEntryBreakpoint methodsFor: 'accessing' stamp: 'AndreiChis 9/7/2017 09:28'!
targetMethod
	^ targetMethod! !

!GTDebuggerMethodEntryBreakpoint methodsFor: 'accessing' stamp: 'AndreiChis 9/7/2017 09:28'!
targetMethod: anObject
	targetMethod := anObject! !

!GTDebuggerMethodEntryBreakpoint methodsFor: 'actions' stamp: 'AndreiChis 9/7/2017 18:25'!
updateDebuggingSession: aSession
	"I remove the reflectivity instrumentation from the top of the stack.
	The at this point the stack in the <suspended context> from the <interrupted process> has this shape: 
	   [1] suspended: MorphicUIManager>>debugProcess:context:label:fullView:notification:usingDebugger:initializedWith:
		[2] sender:    GTPPStepToNextProductionReflectivityAction(GTReflectivityDebugAction)>>openDebuggerForBreakpoint:
		[3] sender:    GTPPStepToNextProductionReflectivityAction(GTReflectivityDebugAction)>>breakpointReachedFrom:
		[4] sender:    GTDebuggerMethodEntryBreakpoint(GTDebuggerBrekpoint)>>breakpointReached
      [5] sender:    <targetMethod> called from the reflective method; this contains the reflectivity metalink. 
	-> [6] sender:    ReflectiveMethod>>run:with:in: -> [this is the context that needs to be removed]
		[7] sender:    method calling the reflective method -> [this needs to call directly <targetMethod>].
		
	At this point the <interrupted context> from the session has at the context [5]."
	super updateDebuggingSession: aSession.
	
	"Step over to ensure that the context that opened the debugger is no longer on the stack"
	aSession stepOver.
	"After this call both the interrupted context and suspended context point to [5]"
	
	"Remote the sender containing the reflectivity method calling the target methodd"
	"The current interrupted context is the target method"
	"The sender is the reflectivity method that calls the method"
	"The sender of the sender is the context calling the method"
	aSession interruptedContext 
		swapSender: aSession interruptedContext sender sender.
		
	aSession interruptedProcess restartTopWith: self targetMethod. "Restart the top context to remove the instrumented code"
	aSession interruptedProcess stepToSendOrReturn. "Find the first sender to simulate how the debugger actions work"
	aSession updateContextTo: aSession interruptedProcess suspendedContext. "Update the session as a normal debugging action"! !

!Object methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 9/17/2013 18:59'!
moldableHalt
	
	MoldableHalt now! !

!Process methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 1/26/2015 23:02'!
debugMoldableHalt: context title: title

	^ (Smalltalk tools debuggerDispatcherFor: context) ifNotNilDo: [ :aDebugger | 
		UIManager default  
			openDebugger: aDebugger
			initializedWith: [  ]
			on: self 
			context: context 
			label: title 
			contents: nil 
			fullView: true ]! !

!Process methodsFor: '*GT-DebuggerExtensions' stamp: 'AndreiChis 8/26/2013 11:51'!
stepUntil: aPredicate from: aContext
	"Resume self until the top context safisfy the given condition."

	| ctxt  |
	
	ctxt := aContext.
	[ 
		ctxt notNil and: [aPredicate matches: ctxt]
	] whileFalse: [
		ctxt := self step: ctxt.
	].
	^ ctxt! !

Halt subclass: #MoldableHalt
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-DebuggerExtensions'!

!MoldableHalt methodsFor: 'priv handling' stamp: 'AndreiChis 9/17/2013 19:01'!
defaultAction

	^ UIManager default handleMoldableHalt: self.! !
